/*
Quadraturdekoder fr Inkrementalgeber fr die neueren ATtiny-AVR0.
Mit einer Codegre von < 900 Byte reicht der kleinste Typ ATtiny202 im 8-pol. Gehuse.
Der Code luft aber auch unverndert auf greren Tinies.

Die phasenverschobenen Signale von Weg- oder Drehgebern PH-A und PH-B werden mit
400 kHz eingelesen und in der ext. ISR mit timer0_comp0a() bewertet.
Zustzlich wird ein /Reset-Eingang bewertet, der den internen Zhler auf 0 setzt.

Der interne 32 Bit Zhler wird per IIC ausgelesen, wobei die Reihenfolge LSB->MSB
per Befehl vertauscht werden kann.

Belegung der I/O-Leitungen:
PA0:  UPDI, Programmieriengang
PA1:  IIC - SDA,  Pullup aktiviert
PA2:  IIC - SCL,  Pullup aktiviert
PA3:  /Reset-Eingang,  Pullup aktiviert
PA6:  PH-A-Eingang,  Pullup aktiviert zustzlich 3k3 empfohlen
PA7:  PH-B-Eingang,  Pullup aktiviert zustzlich 3k3 empfohlen

Die ISR-Funktion erfordert die Verwendung einer aktuellen IDE von IAR fr AVR,
da interne Register reserviert sind und nicht extern benutzt werden drfen.
Es reicht die kostenlose Demoversion: IAR-EW for AVR mit 4 KB Limitierung.

Alle Angaben ohne Gewhr!

Michael Nowak
www.mino-elektronik.de
2020-04-20
*/

#include <stdint.h>
#include <iotiny416.h>
#include <intrinsics.h>

#define ABFRAGERATE   50                // alle 50 Zyklen -> 400 kHz
#define TWI_ADRESSE   0xB8              // default Adresse
#define SCL_PIN    2                    // PA2
#define SDA_PIN    1                    // PA1
#define VERSION   100

extern void init_clear_cnt(void);       // init + lschen
extern void schreibe_cnt(long wert);    // schreiben
extern long lese_cnt(void);             // lesen

#define TX_BUFF_SIZE  4
uint8_t twi_tx_puffer[TX_BUFF_SIZE];
uint8_t twi_tx_index;

uint8_t twi_adresse;
uint8_t twi_status, twi_cmd;

void init_iic_alt()                           // PA1 + PA2 verwenden
{
  PORTA_PIN1CTRL = PORT_PULLUPEN_bm;          // mit pullup
  PORTA_PIN2CTRL = PORT_PULLUPEN_bm;          // mit pullup

  PORTMUX.CTRLB |= PORTMUX_TWI0_bm;						// PA1+PA2 als alternative 
  TWI0_SCTRLA = TWI_ENABLE_bm;
  TWI0_SADDR  = twi_adresse;
}  

// LSByte zuerst ausgeben
void fuelle_puffer(int32_t i)                 // long -> TWI-Puffer
{
  *(uint32_t *)twi_tx_puffer = i;
}  

// Byte Reihenfolge aendern
void invertiere_puffer()
{
uint8_t temp;
  temp = twi_tx_puffer[0];
  twi_tx_puffer[0] = twi_tx_puffer[3];
  twi_tx_puffer[3] = temp;
  temp = twi_tx_puffer[1];
  twi_tx_puffer[1] = twi_tx_puffer[2];
  twi_tx_puffer[2] = temp;
}

void init_qcnt()
{
  PORTA_PIN3CTRL = PORT_PULLUPEN_bm;          // RESET mit pullup
  PORTA_PIN6CTRL = PORT_PULLUPEN_bm;          // PHASE_A mit pullup
  PORTA_PIN7CTRL = PORT_PULLUPEN_bm;          // PHASE_B mit pullup
  
  TCA0.SINGLE.PER = ABFRAGERATE-1;
  TCA0.SINGLE.INTCTRL = TCA_SINGLE_OVF_bm;
  TCA0.SINGLE.CTRLA = TCA_SINGLE_ENABLE_bm;
  init_clear_cnt();                           // implizit:  __enable_interrupt();
}

int main(void)
{
  CCP = CCP_IOREG_gc;                         // IO Register Protection aufheben
  CLKCTRL_MCLKCTRLB = 0x00;                   // kein Vorteiler

  twi_adresse = *(uint8_t *)EEPROM_START;
  if(twi_adresse != 0xff) {
    twi_adresse &= 0xfe;
  } else {
    twi_adresse = TWI_ADRESSE;
  } 
 
  init_iic_alt();                             // alternative Pins fr SDA+SCL
  init_qcnt();                                // 
  while(1) {
    if(TWI0_SSTATUS & (TWI_DIF_bm | TWI_APIEN_bm)) { // Adr. oder Byte empfangen
      twi_status = TWI0_SSTATUS;	
      if(twi_status & TWI_APIF_bm) {          // adresse empfangen
        if (!(TWI0_SSTATUS & TWI_AP_bm)) {    // bei stop condition
          TWI0_SSTATUS |= TWI_APIF_bm;        // alle Flags lschen
        } else {
          TWI0_SCTRLB = TWI_SCMD_RESPONSE_gc; // ack ausgeben
        }  
      }
      else if(twi_status & TWI_DIF_bm) {
        if(!(TWI0_SSTATUS&TWI_DIR_bm)) {      // Daten vom master empfangen
          twi_cmd = TWI0_SDATA;               // 1. Byte = Befehl
          TWI0_SCTRLB = TWI_SCMD_RESPONSE_gc; // ack ausgeben
          twi_tx_index = 0;                   // Sendepuffer auf Anfang
          if(twi_cmd == 'V') fuelle_puffer(VERSION);
          else if(twi_cmd == 'A') fuelle_puffer(lese_cnt());
          else if(twi_cmd == 'B') fuelle_puffer(lese_cnt()/2);
          else if(twi_cmd == 'R') invertiere_puffer(); // MSB zuerst ausgeben
          else if(twi_cmd == 'Z') init_clear_cnt(); // 'Z' = Reset
        } else {                              // master will lesen  
          TWI0_SDATA = twi_tx_puffer[twi_tx_index++];  // Byte aus Ringpuffer senden
          twi_tx_index %= TX_BUFF_SIZE;
          TWI0_SCTRLB = TWI_SCMD_RESPONSE_gc; // ack ausgeben
        }  
      }
    } 
  }
}